@@ -96,7 +96,7 @@ static bool s_parse_expiration_value_from_json_object(
             if (aws_json_value_get_string(value, &expiration_cursor)) {
                 AWS_LOGF_INFO(
                     AWS_LS_AUTH_CREDENTIALS_PROVIDER,
-                    "Unabled to extract credentials Expiration field from Json document.");
+                    "Unable to extract credentials Expiration field from Json document.");
                 return false;
             }
 
@@ -123,7 +123,7 @@ static bool s_parse_expiration_value_from_json_object(
             if (aws_json_value_get_number(value, &expiration_value)) {
                 AWS_LOGF_INFO(
                     AWS_LS_AUTH_CREDENTIALS_PROVIDER,
-                    "Unabled to extract credentials Expiration field from Json document.");
+                    "Unable to extract credentials Expiration field from Json document.");
                 return false;
             }
 
@@ -136,7 +136,7 @@ static bool s_parse_expiration_value_from_json_object(
             if (aws_json_value_get_number(value, &expiration_value_ms)) {
                 AWS_LOGF_INFO(
                     AWS_LS_AUTH_CREDENTIALS_PROVIDER,
-                    "Unabled to extract credentials Expiration field from Json document.");
+                    "Unable to extract credentials Expiration field from Json document.");
                 return false;
             }
 
@@ -171,8 +171,9 @@ struct aws_credentials *aws_parse_credentials_from_aws_json_object(
 
     struct aws_credentials *credentials = NULL;
     struct aws_json_value *access_key_id = NULL;
-    struct aws_json_value *secrete_access_key = NULL;
+    struct aws_json_value *secret_access_key = NULL;
     struct aws_json_value *token = NULL;
+    struct aws_json_value *account_id = NULL;
     struct aws_json_value *creds_expiration = NULL;
 
     bool parse_error = true;
@@ -189,19 +190,21 @@ struct aws_credentials *aws_parse_credentials_from_aws_json_object(
         goto done;
     }
 
-    struct aws_byte_cursor secrete_access_key_cursor;
-    secrete_access_key = aws_json_value_get_from_object(
+    struct aws_byte_cursor secret_access_key_cursor;
+    secret_access_key = aws_json_value_get_from_object(
         document_root, aws_byte_cursor_from_c_str((char *)options->secret_access_key_name));
-    if (!aws_json_value_is_string(secrete_access_key) ||
-        aws_json_value_get_string(secrete_access_key, &secrete_access_key_cursor) == AWS_OP_ERR) {
+    if (!aws_json_value_is_string(secret_access_key) ||
+        aws_json_value_get_string(secret_access_key, &secret_access_key_cursor) == AWS_OP_ERR) {
         AWS_LOGF_ERROR(AWS_LS_AUTH_CREDENTIALS_PROVIDER, "Failed to parse SecretAccessKey from Json document.");
         goto done;
     }
 
     struct aws_byte_cursor token_cursor;
+    AWS_ZERO_STRUCT(token_cursor);
     if (options->token_name) {
         token = aws_json_value_get_from_object(document_root, aws_byte_cursor_from_c_str((char *)options->token_name));
-        if (!aws_json_value_is_string(token) || aws_json_value_get_string(token, &token_cursor) == AWS_OP_ERR) {
+        if (!aws_json_value_is_string(token) || aws_json_value_get_string(token, &token_cursor) == AWS_OP_ERR ||
+            token_cursor.len == 0) {
             if (options->token_required) {
                 AWS_LOGF_ERROR(AWS_LS_AUTH_CREDENTIALS_PROVIDER, "Failed to parse Token from Json document.");
                 goto done;
@@ -209,6 +212,17 @@ struct aws_credentials *aws_parse_credentials_from_aws_json_object(
         }
     }
 
+    struct aws_byte_cursor account_id_cursor;
+    AWS_ZERO_STRUCT(account_id_cursor);
+    if (options->account_id_name) {
+        account_id =
+            aws_json_value_get_from_object(document_root, aws_byte_cursor_from_c_str((char *)options->account_id_name));
+        if (!aws_json_value_is_string(account_id) ||
+            aws_json_value_get_string(account_id, &account_id_cursor) == AWS_OP_ERR) {
+            AWS_LOGF_DEBUG(AWS_LS_AUTH_CREDENTIALS_PROVIDER, "Failed to parse account_id from Json document.");
+        }
+    }
+
     // needed to avoid uninitialized local variable error
     uint64_t expiration_timepoint_in_seconds = UINT64_MAX;
     if (options->expiration_name) {
@@ -226,31 +240,20 @@ struct aws_credentials *aws_parse_credentials_from_aws_json_object(
     /*
      * Build the credentials
      */
-    if (access_key_id_cursor.len == 0 || secrete_access_key_cursor.len == 0) {
+    if (access_key_id_cursor.len == 0 || secret_access_key_cursor.len == 0) {
         AWS_LOGF_ERROR(
             AWS_LS_AUTH_CREDENTIALS_PROVIDER,
             "Parsed an unexpected credentials json document, either access key, secret key is empty.");
         goto done;
     }
-
-    struct aws_byte_cursor session_token_cursor;
-    AWS_ZERO_STRUCT(session_token_cursor);
-
-    if (token) {
-        aws_json_value_get_string(token, &session_token_cursor);
-        if (options->token_required && session_token_cursor.len == 0) {
-            AWS_LOGF_ERROR(
-                AWS_LS_AUTH_CREDENTIALS_PROVIDER, "Parsed an unexpected credentials json document with empty token.");
-            goto done;
-        }
-    }
-
-    credentials = aws_credentials_new(
-        allocator,
-        access_key_id_cursor,
-        secrete_access_key_cursor,
-        session_token_cursor,
-        expiration_timepoint_in_seconds);
+    struct aws_credentials_options creds_option = {
+        .access_key_id_cursor = access_key_id_cursor,
+        .secret_access_key_cursor = secret_access_key_cursor,
+        .session_token_cursor = token_cursor,
+        .account_id_cursor = account_id_cursor,
+        .expiration_timepoint_seconds = expiration_timepoint_in_seconds,
+    };
+    credentials = aws_credentials_new_with_options(allocator, &creds_option);
 
     if (credentials == NULL) {
         AWS_LOGF_ERROR(AWS_LS_AUTH_CREDENTIALS_PROVIDER, "Failed to allocate memory for credentials.");
@@ -383,15 +386,113 @@ static struct aws_byte_cursor s_aws_isoe_dns_suffix = AWS_BYTE_CUR_INIT_FROM_STR
 static struct aws_byte_cursor s_isof_region_prefix = AWS_BYTE_CUR_INIT_FROM_STRING_LITERAL("us-isof-");
 static struct aws_byte_cursor s_aws_isof_dns_suffix = AWS_BYTE_CUR_INIT_FROM_STRING_LITERAL("csp.hci.ic.gov");
 
-int aws_credentials_provider_construct_regional_endpoint(
+AWS_STATIC_STRING_FROM_LITERAL(s_endpoint_url_env, "AWS_ENDPOINT_URL");
+AWS_STATIC_STRING_FROM_LITERAL(s_endpoint_url_property, "endpoint_url");
+AWS_STATIC_STRING_FROM_LITERAL(s_services_property, "services");
+
+/*
+ * Resolve any endpoint overrides for the service
+ * See: https://docs.aws.amazon.com/sdkref/latest/guide/feature-ss-endpoints.html
+ *
+ * The order of resolution for configured endpoint is as follows:
+ * 1. The value provided by a service-specific environment variable, `AWS_ENDPOINT_URL_<SERVICE>`.
+ * 2. The value provided by the global endpoint environment variable, `AWS_ENDPOINT_URL`.
+ * 3. The value provided by a service-specific parameter from a services definition section referenced in a profile.
+ * 4. The value provided by the global parameter from a profile in the shared configuration file.
+ *
+ * Returns NULL if there are no endpoint overrides configured
+ */
+struct aws_string *s_get_override_endpoint(
+    struct aws_allocator *allocator,
+    const struct aws_string *override_service_name_env,
+    const struct aws_string *override_service_name_property,
+    const struct aws_profile_collection *profile_collection,
+    const struct aws_profile *profile) {
+
+    struct aws_byte_cursor url_env_cursor = aws_byte_cursor_from_string(s_endpoint_url_env);
+    struct aws_byte_cursor underscore_cursor = aws_byte_cursor_from_c_str("_");
+    struct aws_byte_cursor service_name_env_cursor = aws_byte_cursor_from_string(override_service_name_env);
+
+    struct aws_string *out_endpoint = NULL;
+    struct aws_string *service_endpoint_str = NULL;
+
+    struct aws_byte_buf service_endpoint_buf;
+    AWS_ZERO_STRUCT(service_endpoint_buf);
+    /* Check AWS_ENDPOINT_URL_SERVICENAME variable */
+    aws_byte_buf_init(&service_endpoint_buf, allocator, 10);
+    if (aws_byte_buf_append_dynamic(&service_endpoint_buf, &url_env_cursor) ||
+        aws_byte_buf_append_dynamic(&service_endpoint_buf, &underscore_cursor) ||
+        aws_byte_buf_append_dynamic(&service_endpoint_buf, &service_name_env_cursor)) {
+        goto on_finish;
+    }
+    service_endpoint_str = aws_string_new_from_buf(allocator, &service_endpoint_buf);
+    out_endpoint = aws_get_env_nonempty(allocator, aws_string_c_str(service_endpoint_str));
+    if (out_endpoint) {
+        goto on_finish;
+    }
+
+    /* Check AWS_ENDPOINT_URL variable */
+    out_endpoint = aws_get_env_nonempty(allocator, aws_string_c_str(s_endpoint_url_env));
+    if (out_endpoint) {
+        goto on_finish;
+    }
+
+    /* parse endpoint override from config file */
+    if (profile_collection == NULL || profile == NULL) {
+        goto on_finish;
+    }
+
+    /* check services specific endpoint property */
+    const struct aws_profile_property *property = aws_profile_get_property(profile, s_services_property);
+    if (property) {
+        const struct aws_profile *services = aws_profile_collection_get_section(
+            profile_collection, AWS_PROFILE_SECTION_TYPE_SERVICES, aws_profile_property_get_value(property));
+        if (services) {
+            property = aws_profile_get_property(services, override_service_name_property);
+            if (property) {
+                out_endpoint = aws_string_new_from_string(
+                    allocator, aws_profile_property_get_sub_property(property, s_endpoint_url_property));
+                goto on_finish;
+            }
+        }
+        goto on_finish;
+    }
+
+    /* check aws_endpoint_url property */
+    property = aws_profile_get_property(profile, s_endpoint_url_property);
+    if (property) {
+        out_endpoint = aws_string_new_from_string(allocator, aws_profile_property_get_value(property));
+
+        goto on_finish;
+    }
+
+on_finish:
+    aws_byte_buf_clean_up(&service_endpoint_buf);
+    aws_string_destroy(service_endpoint_str);
+    return out_endpoint;
+}
+
+int aws_credentials_provider_construct_endpoint(
     struct aws_allocator *allocator,
     struct aws_string **out_endpoint,
     const struct aws_string *region,
-    const struct aws_string *service_name) {
+    const struct aws_string *service_name_host,
+    const struct aws_string *service_name_env,
+    const struct aws_string *service_name_property,
+    const struct aws_profile_collection *profile_collection,
+    const struct aws_profile *profile,
+    const struct aws_byte_cursor *suffix_override,
+    bool swap_region_and_service_name) {
+
+    *out_endpoint =
+        s_get_override_endpoint(allocator, service_name_env, service_name_property, profile_collection, profile);
+    if (*out_endpoint) {
+        return AWS_OP_SUCCESS;
+    }
 
     AWS_PRECONDITION(allocator);
     AWS_PRECONDITION(out_endpoint);
-    if (!region || !service_name) {
+    if (!region || !service_name_host) {
         return aws_raise_error(AWS_ERROR_INVALID_ARGUMENT);
     }
     int result = AWS_OP_ERR;
@@ -399,19 +500,32 @@ int aws_credentials_provider_construct_regional_endpoint(
     struct aws_byte_buf endpoint;
     AWS_ZERO_STRUCT(endpoint);
     aws_byte_buf_init(&endpoint, allocator, 10);
-    struct aws_byte_cursor service_cursor = aws_byte_cursor_from_string(service_name);
+    struct aws_byte_cursor service_cursor = aws_byte_cursor_from_string(service_name_host);
     struct aws_byte_cursor region_cursor = aws_byte_cursor_from_string(region);
 
-    if (aws_byte_buf_append_dynamic(&endpoint, &service_cursor) ||
-        aws_byte_buf_append_dynamic(&endpoint, &s_dot_cursor) ||
-        aws_byte_buf_append_dynamic(&endpoint, &region_cursor) ||
-        aws_byte_buf_append_dynamic(&endpoint, &s_dot_cursor)) {
-        goto on_error;
+    if (swap_region_and_service_name) {
+        if (aws_byte_buf_append_dynamic(&endpoint, &region_cursor) ||
+            aws_byte_buf_append_dynamic(&endpoint, &s_dot_cursor) ||
+            aws_byte_buf_append_dynamic(&endpoint, &service_cursor) ||
+            aws_byte_buf_append_dynamic(&endpoint, &s_dot_cursor)) {
+            goto on_error;
+        }
+    } else {
+        if (aws_byte_buf_append_dynamic(&endpoint, &service_cursor) ||
+            aws_byte_buf_append_dynamic(&endpoint, &s_dot_cursor) ||
+            aws_byte_buf_append_dynamic(&endpoint, &region_cursor) ||
+            aws_byte_buf_append_dynamic(&endpoint, &s_dot_cursor)) {
+            goto on_error;
+        }
     }
 
     const struct aws_byte_cursor region_cur = aws_byte_cursor_from_string(region);
 
-    if (aws_byte_cursor_starts_with(&region_cur, &s_cn_region_prefix)) { /* AWS CN partition */
+    if (suffix_override) {
+        if (aws_byte_buf_append_dynamic(&endpoint, suffix_override)) {
+            goto on_error;
+        }
+    } else if (aws_byte_cursor_starts_with(&region_cur, &s_cn_region_prefix)) { /* AWS CN partition */
         if (aws_byte_buf_append_dynamic(&endpoint, &s_aws_cn_dns_suffix)) {
             goto on_error;
         }
@@ -452,14 +566,31 @@ AWS_STATIC_STRING_FROM_LITERAL(s_region_env, "AWS_REGION");
 AWS_STATIC_STRING_FROM_LITERAL(s_default_region_env, "AWS_DEFAULT_REGION");
 
 struct aws_string *aws_credentials_provider_resolve_region_from_env(struct aws_allocator *allocator) {
-    struct aws_string *region = NULL;
-
     /* check AWS_REGION environment variable first */
-    aws_get_environment_value(allocator, s_region_env, &region);
-    if (region != NULL && region->len > 0) {
+    struct aws_string *region = aws_get_env_nonempty(allocator, aws_string_c_str(s_region_env));
+    if (region != NULL) {
         return region;
     }
 
-    aws_get_environment_value(allocator, s_default_region_env, &region);
-    return region;
+    /* check AWS_DEFAULT_REGION environment variable first */
+    return aws_get_env_nonempty(allocator, aws_string_c_str(s_default_region_env));
+}
+
+struct aws_byte_cursor aws_parse_account_id_from_arn(struct aws_byte_cursor arn) {
+    struct aws_byte_cursor account_id;
+    AWS_ZERO_STRUCT(account_id);
+    /* The format of the Arn is arn:partition:service:region:account-id:resource-ID and we need to parse the
+     * account-id out of it which is the fifth element. */
+    for (int i = 0; i < 5; i++) {
+        if (!aws_byte_cursor_next_split(&arn, ':', &account_id)) {
+            AWS_LOGF_ERROR(
+                AWS_LS_AUTH_CREDENTIALS_PROVIDER,
+                "Failed to parse account_id string from STS xml response: %s",
+                aws_error_str(aws_last_error()));
+            struct aws_byte_cursor empty;
+            AWS_ZERO_STRUCT(empty);
+            return empty;
+        }
+    }
+    return account_id;
 }
